package flink.zanxiangnet.ad.monitoring.process;

import flink.zanxiangnet.ad.monitoring.config.MysqlDataSourceFactory;
import flink.zanxiangnet.ad.monitoring.dao.mapper.AdStatOfDayDWDMapper;
import flink.zanxiangnet.ad.monitoring.pojo.entity.*;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.util.DateUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.flink.api.common.state.*;
import org.apache.flink.api.common.typeinfo.Types;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.windows.TimeWindow;
import org.apache.flink.util.Collector;
import org.apache.ibatis.datasource.DataSourceFactory;
import org.apache.ibatis.mapping.Environment;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.ibatis.session.SqlSessionFactoryBuilder;
import org.apache.ibatis.transaction.jdbc.JdbcTransactionFactory;
import org.springframework.beans.BeanUtils;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.stream.Collectors;

@Slf4j
public class AdMinuteDWDProcess extends ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow> {
    private static final DateTimeFormatter formatForLastReduceKey = DateTimeFormatter.ofPattern("yyyy-MM-dd_HH:mm");

    private SqlSessionFactory sqlSessionFactory;
    // 上次查询的天数据的时间
    private ValueState<Long> lastQueryTimeState;
    // 历史的天数据
    private ListState<AdStatOfDayDWD> historyDayState;
    // 前 5分钟聚合的数据
    private MapState<String, AdStatOfMinuteDWD> lastReduceState;

    public AdMinuteDWDProcess() {
    }

    @Override
    public void open(Configuration conf) {
        Map<String, String> params = getRuntimeContext()
                .getExecutionConfig()
                .getGlobalJobParameters()
                .toMap();

        Properties mysqlProps = new Properties();
        mysqlProps.setProperty(ApplicationProperties.MYSQL_DRIVER_CLASS_NAME, params.get(ApplicationProperties.MYSQL_DRIVER_CLASS_NAME));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_URL, params.get(ApplicationProperties.MYSQL_URL));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_USERNAME, params.get(ApplicationProperties.MYSQL_USERNAME));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_PASSWORD, params.get(ApplicationProperties.MYSQL_PASSWORD));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_INIT_POOL_SIZE, params.get(ApplicationProperties.MYSQL_INIT_POOL_SIZE));
        mysqlProps.setProperty(ApplicationProperties.MYSQL_MAX_POOL_SIZE, params.get(ApplicationProperties.MYSQL_MAX_POOL_SIZE));

        DataSourceFactory dataSourceFactory = new MysqlDataSourceFactory();
        dataSourceFactory.setProperties(mysqlProps);
        Environment environment = new Environment("mysql", new JdbcTransactionFactory(), dataSourceFactory.getDataSource());
        org.apache.ibatis.session.Configuration configuration = new org.apache.ibatis.session.Configuration(environment);
        // 开启驼峰规则
        configuration.setMapUnderscoreToCamelCase(true);
        configuration.getTypeAliasRegistry().registerAlias(AdStatOfDayDWD.class);
        // addMapper一定要放到 alias的后面！！！！！
        configuration.addMapper(AdStatOfDayDWDMapper.class);
        sqlSessionFactory = new SqlSessionFactoryBuilder().build(configuration);

        historyDayState = getRuntimeContext().getListState(new ListStateDescriptor<>("historyDayState", Types.POJO(AdStatOfDayDWD.class)));
        lastQueryTimeState = getRuntimeContext().getState(new ValueStateDescriptor<>("lastQueryTimeState", Types.LONG));
        lastReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("lastReduceState", Types.STRING, Types.POJO(AdStatOfMinuteDWD.class)));
    }

    @Override
    public void process(Long elementCount, ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow>.Context context,
                        Iterable<AdDataOfMinuteODS> iterable, Collector<AdStatOfMinuteDWD> collector) throws Exception {
        long beginTime = context.window().getStart();
        LocalDateTime beginDateTime = DateUtil.milliToLocalDateTime(beginTime);
        LocalDate beginDate = beginDateTime.toLocalDate();
        String statDay = DateUtil.formatLocalDate(beginDate);
        int hour = beginDateTime.getHour();
        long now = System.currentTimeMillis();

        AdDataOfMinuteODS thisHourMinuteODS = null;
        List<AdDataOfMinuteODS> thisHourMinuteODSList = new ArrayList<>(24);
        for (AdDataOfMinuteODS minuteODS : iterable) {
            thisHourMinuteODSList.add(minuteODS);
            if (minuteODS.getHour() == hour) {
                thisHourMinuteODS = minuteODS;
            }
        }
        Long adId = thisHourMinuteODSList.get(0).getAdId();

        // 每 2小时更新下之前的历史数据
        Long lastQueryTime = lastQueryTimeState.value();
        if (lastQueryTime == null || (beginTime - lastQueryTime > 2 * 60 * 60 * 1000L)) {
            try (SqlSession session = sqlSessionFactory.openSession()) {
                AdStatOfDayDWDMapper mapper = session.getMapper(AdStatOfDayDWDMapper.class);
                List<AdStatOfDayDWD> historyData = mapper.lastReduceResult(adId, null, DateUtil.formatLocalDate(beginDate.minusDays(1L)), 6);
                historyDayState.clear();
                if (!historyData.isEmpty()) {
                    historyDayState.update(historyData);
                }
                lastQueryTimeState.update(beginTime);
            }
        }

        AdStatOfDayDWD yesterdayDayDWD = null;
        for (AdStatOfDayDWD dayDWD : historyDayState.get()) {
            if (dayDWD.getStatDay().compareTo(statDay) < 0) {
                yesterdayDayDWD = dayDWD;
                break;
            }
        }

        if (thisHourMinuteODS == null && hour > 0) {
            int minHour = hour;
            AdDataOfMinuteODS lastHourMinuteODS = null;
            List<AdDataOfMinuteODS> lastHourMinuteODSList = new ArrayList<>(23);
            for (AdDataOfMinuteODS minuteODS : thisHourMinuteODSList) {
                if (minuteODS.getHour() >= hour) {
                    continue;
                }
                minHour = Math.min(minHour, minuteODS.getHour());
                lastHourMinuteODSList.add(minuteODS);
                if (lastHourMinuteODS == null) {
                    lastHourMinuteODS = minuteODS;
                } else {
                    lastHourMinuteODS = minuteODS.getHour() > lastHourMinuteODS.getHour() ? minuteODS : lastHourMinuteODS;
                }
            }
            if (lastHourMinuteODS != null) {
                if (lastHourMinuteODS.getHour() == hour - 1) {
                    // 重新聚合上小时的数据
                    if (beginDateTime.getMinute() == 0 || beginDateTime.getMinute() == 15 || beginDateTime.getMinute() == 55) {
                        LocalDateTime lastHourStatTime = LocalDateTime.of(beginDate, LocalTime.of(hour - 1, 55, 0));
                        lastHourMinuteODS.setStatTime(DateUtil.localDateTimeToMilli(lastHourStatTime));
                        AdStatOfMinuteDWD lastReduce = lastReduceState.get(lastHourStatTime.minusMinutes(5L).format(formatForLastReduceKey));
                        // 聚合当天的全部数据
                        AdStatOfMinuteDWD lastHourAdStat = AdStatOfMinuteDWD.reduce(yesterdayDayDWD, lastHourMinuteODSList, lastHourMinuteODS, lastReduce, now);
                        collector.collect(lastHourAdStat);
                    }
                }
                // 造一条空数据
                if (beginDateTime.getMinute() == 0 && minHour == lastHourMinuteODS.getHour()) {
                    LocalDateTime hourStatTime = LocalDateTime.of(beginDate, LocalTime.of(hour, 0, 0));
                    // 找今天聚合的最后一条历史数据
                    AdStatOfMinuteDWD temp = null;
                    for (int i = hour - 1; i >= 0; i--) {
                        temp = lastReduceState.get(LocalDateTime.of(beginDate, LocalTime.of(i, 55, 0)).format(formatForLastReduceKey));
                        if (temp != null) {
                            break;
                        }
                    }
                    if (temp != null) {
                        AdStatOfMinuteDWD hourTemp = new AdStatOfMinuteDWD();
                        BeanUtils.copyProperties(temp, hourTemp);
                        hourTemp.setHour(hour);
                        hourTemp.setStatTime(DateUtil.localDateTimeToMilli(hourStatTime));
                        AdStatOfMinuteDWD.resetHourAndMinute(hourTemp);
                        collector.collect(hourTemp);
                    } else {
                        log.error("造数据时没有找到历史数据, adId: {}, hour: {}, minHour: {}", adId, hour, minHour);
                    }
                }
            }
        }


        if (thisHourMinuteODS != null) {
            AdStatOfMinuteDWD lastReduce = lastReduceState.get(beginDateTime.minusMinutes(5L).format(formatForLastReduceKey));
            // 聚合当天的全部数据
            AdStatOfMinuteDWD newAdStat = AdStatOfMinuteDWD.reduce(yesterdayDayDWD, thisHourMinuteODSList, thisHourMinuteODS, lastReduce, now);
            collector.collect(newAdStat);
            lastReduceState.put(beginDateTime.format(formatForLastReduceKey), newAdStat);
        }

        clearState(beginDateTime);
    }

    @Override
    public void clear(ProcessWindowFunction<AdDataOfMinuteODS, AdStatOfMinuteDWD, Long, TimeWindow>.Context context) throws Exception {
    }

    private void clearState(LocalDateTime beginDateTime) throws Exception {
        List<String> removeKeys = new ArrayList<>(20);
        Iterator<Map.Entry<String, AdStatOfMinuteDWD>> lastIterator = lastReduceState.iterator();

        String delLastReduceKey = beginDateTime.minusDays(1).format(formatForLastReduceKey);
        while (lastIterator.hasNext()) {
            Map.Entry<String, AdStatOfMinuteDWD> temp = lastIterator.next();
            if (temp.getKey().compareTo(delLastReduceKey) < 0) {
                removeKeys.add(temp.getKey());
            }
        }
        if (!removeKeys.isEmpty()) {
            for (String key : removeKeys) {
                lastReduceState.remove(key);
            }
        }
    }
}
